# LSB Test Execution Framework
# Packages management module
#
# Copyright (C) 2007 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Packages;
use strict;

use POSIX;

require Exporter;
@Packages::ISA = qw(Exporter);
# Exported symbols
push @Packages::EXPORT, qw(
    &install_package &list_files &print_files_list 
    &guess_platform &package_info &remove_package 
    &prepare_test &install_test 
    &guess_package_manager &inquire_for_files_list
    $supported_package_managers @all_platforms @all_lsb_versions
    &packages_info &package_file_location
    $FTP_srv $FTP_lsb
);

use Common;

# This is the primary package source.
our $FTP_srv = "ftp.linux-foundation.org";
our $FTP_lsb = "ftp://$FTP_srv/pub/lsb";
# Use '--ftp <server>/dir' option to change this. 

# FTP handler
my $ftp;

# Mode: i (installed) - package is installed
#       p (present) - file presents in the local package_dir
#       a (available) - file is available for downloading from the ftp source.
#       b (beta) - file is available for downloading from beta source.

# List of all available files. src = "" means local package_dir. 
# filesize is the size to download, it equals 0 for local files.
my $s_files = {}; # $s_files->{$mode}->{lsb_ver}->{arch}->{package_name}->
                      # -> %file = ( FILENAME=>qm-2.2-2.ia32, 
                                   # PACKAGE=>qm, PKG_VER=>2.2-2, MODE=>[ipab],
                                   # SRC=>src, SIZE=>size_to_dwnld, 
                                   # LSB_VER=>3.1, ARCH=>ia32  );

my %files = (); # $files{package_name.version.arch}->%file

# Information about required files for each certain 
# test and architecture
my $test_info = {};    # $test_info->{lsb_ver}->{arch}->{test_name}->
                           # ->{mode}->{FILES}->{package_name}->%file
                           #        \->{VERSION}

# List of all platform names. 
our @all_platforms = ("amd64", "ia32", "ia64", "ppc32", "ppc64", "s390", "s390x");
our @all_lsb_versions = ("3.0", "3.1", "3.2", "snapshot");

our $supported_package_managers = "RPM and dpkg(via alien)";

# Expiration time for cached FTP listings (in seconds).
my $expire_period = 60*60*24*7; # 7 days
my $oldest_timestamp = undef;
#------------------------------------------------------------------------

# Helps to determine platform name.
sub guess_platform 
{
    my $name = shift;
    
    if ( $name =~ /^(amd64|x86_64)$/)
        { return "amd64" }
    elsif ( $name =~ /^(ia32|i[3-6]86|x86)$/ )
        { return "ia32" }
    elsif ( $name =~ /^(ia64)$/ )
        { return "ia64" }
    elsif ( $name =~ /^(ppc|ppc32)$/ )
        { return "ppc32" }
    elsif ( $name =~ /^(ppc64)$/ )
        { return "ppc64" }
    elsif ( $name =~ /^(s390)$/ )
        { return "s390" }
    elsif ( $name =~ /^(s390x)$/ )
        { return "s390x" }
    else
        { return "unknown"; }
}
#------------------------------------------------------------------------

# Returns 1 if $platform is backward compatible with the host_platform,
# 0 otherwise.
sub is_backward_compatible_platform {
    my $platform = shift;
    
    if ( $platform eq $host_platform ) {
        return 1;
    }
    
    if ( $host_platform eq "amd64" ) {
        if ( $platform eq "ia32" ) {
            return 1;
        }
    } elsif ( $host_platform eq "ia64" ) {
        if ( $platform eq "ia32" ) {
            return 1;
        }
    } elsif ( $host_platform eq "ppc64" ) {
        if ( $platform eq "ppc32" ) {
            return 1;
        }
    } elsif ( $host_platform eq "s390x" ) {
        if ( $platform eq "s390" ) {
            return 1;
        }
    }
    
    return 0;
}

#------------------------------------------------------------------------

# Determines which package manager to use.
sub guess_package_manager {
    debug_inform("Host OS: $host_os");
    if ( $host_os =~ /(suse|red\W?hat|asianux|mandriva|turbolinux|gentoo)/i )
    {
        # Native package manager for Suse and Redhat is RPM
        return "RPM";
    }
    elsif ( $host_os =~ /(ubuntu|debian|freespire|linspire|rays.?lx)/i )
    {
        # For Debian and Ubuntu use alien and dpkg.
        return "dpkg";
    }
    else {
        if ( does_shell_know ("dpkg") ) {
            return "dpkg";
        } else {
            return "RPM";
        }
    }
    
    return "unknown";
}

#---------------------------------------------------------------------
sub rpm_package_info {
    my $package_name = shift;
    
    my $result = "";
    $result = capture("rpm -q $package_name"); 
    chomp $result; debug_inform( "| $result");
    if ( $? != 0 ) { $result = ""; }
    $result =~ s/^$package_name-//; # get the version
    $result =~ s/\.lsb\d\d?//;      # remove lsb(2|20|3) from version
    $result =~ m/\.([^\.]+)$/;
    if ($1 and (guess_platform($1) ne "unknown")) {
    	$result =~ s/\.[^\.]+$//;
    }
    
    return $result;
}

#---------------------------------------------------------------------

sub dpkg_package_info {
    my $package_name = shift;

    my $result = "";
    debug_inform( "dpkg -s $package_name");
    $result = `dpkg -s $package_name  2>/dev/null | grep '^\\(Version\\|Status\\):'`;
    chomp $result; debug_inform( "dpkg: $?| $result");
        
    if ( $? == 0 ) {
        if ($result !~ m/Status:\s+\w+\s+ok\s+installed/) { $? = 1; }
    }
    
    if ( $? == 0 ) {
        if ($result =~ m/Version:\s+([^\n]+)/) {
            $result = $1;
        } else {
            fail("Can't understand the answer of dpkg: $result"); 
        }
    }
    if ( $? != 0 ) { $result = ""; }
    $result =~ s/\.lsb\d\d?//; # remove lsb(2|20|3)
    
    return $result;
}

#---------------------------------------------------------------------

# Returns info about a package 
# and empty string if the package isn't installed.
sub package_info {
    my $package_name = shift;
    my $true_name_ref = shift;
    my $null;
    
    unless ($true_name_ref) { $true_name_ref = \$null; }
    $package_name =~ s/(^\s+|\s+$)//g; #trim
    $$true_name_ref = $package_name;
    
    my $result = "";
    
    if ( $host_package_manager eq "RPM" )
    {
        $result = rpm_package_info ($package_name);
        if ( !$result && $package_name eq "lsb-tet3" ) {
            $$true_name_ref = "lsb-tet3-lite";
            $result = rpm_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-qm" ) {
            $$true_name_ref = "qm";
            $result = rpm_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-test-libstdcpp" ) {
            $$true_name_ref = "qmtest_libstdcpp";
            $result = rpm_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-test-core" ) {
            $$true_name_ref = "lsb-runtime-test";
            $result = rpm_package_info ($$true_name_ref);
        }
        if ( !$result ) { $$true_name_ref = $package_name; }
    }
    elsif ( $host_package_manager eq "dpkg" )
    {
        # Use 'alien' and 'dpkg'.
        $result = dpkg_package_info ($package_name);
        if ( !$result && $package_name eq "lsb-tet3" ) {
            $$true_name_ref = "lsb-tet3-lite";
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-qm" ) {
            $$true_name_ref = "qm";
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-test-libstdcpp" ) {
            $$true_name_ref = "qmtest-libstdcpp";
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "qmtest_libstdcpp" ) {
            $$true_name_ref = "qmtest-libstdcpp";
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result && $package_name eq "lsb-test-core" ) {
            $$true_name_ref = "lsb-runtime-test";
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result ) {
            $package_name =~ s/_/-/g;
            $$true_name_ref = $package_name;
            $result = dpkg_package_info ($$true_name_ref);
        }
        if ( !$result ) { $$true_name_ref = $package_name; }
    }
    else {
        # Unknown package manager
        fail( "package_info: Unknown package manager."
             ." ($host_package_manager FOR $host_os)"
            ); 
    }
    
    return $result;
}
#------------------------------------------------------------------------

# Returns info about packages in a hash
sub packages_info {
    my @package_names = @_;
    
    my %result = (); # {package_name}->package_version
    
    if ( $host_package_manager eq "RPM" )
    {
        # Use RPM
        for ( @package_names ) {
            if (/^lsb-tet3$/) {
                push @package_names, "lsb-tet3-lite";
            }
            if (/^lsb-qm$/) {
                push @package_names, "qm";
            }
            elsif (/^lsb-test-libstdcpp$/) {
                push @package_names, "qmtest_libstdcpp";
            }
            elsif (/^lsb-test-core$/) {
                push @package_names, "lsb-runtime-test";
            }
        }
        my $resp = capture( "rpm -q ".(join " ",@package_names)." 2>/dev/null" ); 
        chomp $resp; debug_inform( $resp );
        my @lines = split(/\n/, $resp);
        foreach my $line (@lines) {
            my $package_name = shift @package_names;
            # get the version
            if ( $line =~ m/^$package_name-(.*)$/ ) {
                my $ver = $1;
                debug_inform("$package_name: $ver");
                $ver =~ s/\.lsb\d\d?//; # remove .lsb(2|20|3)
		$ver =~ m/\.([^\.]+)$/;
		if( guess_platform($1) ne "unknown") {
		    $ver =~ s/\.[^\.]+$//;
		}
                if ( $package_name eq "lsb-tet3-lite" ) { $package_name = "lsb-tet3"; }
                if ( $package_name eq "qm" ) { $package_name = "lsb-qm"; }
                if ( $package_name eq "qmtest_libstdcpp" ) { $package_name = "lsb-test-libstdcpp"; }
                if ( $package_name eq "lsb-runtime-test" ) { $package_name = "lsb-test-core"; }
                $result{$package_name} = $ver;
            }
        }
    }
    elsif ( $host_package_manager eq "dpkg" )
    {
        # Use 'alien' and 'dpkg'.
        for ( @package_names ) {
            if (/_/) { s/_/-/g; }
            if (/^lsb-tet3$/) {
                push @package_names, "lsb-tet3-lite";
            }
            if (/^lsb-qm$/) {
                push @package_names, "qm";
            }
            elsif (/^lsb-test-libstdcpp$/) {
                push @package_names, "qmtest-libstdcpp";
            }
            elsif (/^lsb-test-core$/) {
                push @package_names, "lsb-runtime-test";
            }
        }
        my $resp = capture( "dpkg -l ".(join " ",@package_names)." 2>/dev/null" );
        chomp $resp; debug_inform( $resp );
        
        my @lines = split(/\n/, $resp);
        foreach my $line (@lines) {
            if ( $line =~ /^\wi\s\s+([^\s]+)\s+([^\s]+)/ ) {
                my $package_name = $1;
                my $package_ver = $2;
                next if ( !in_array($package_name, \@package_names) );
                $package_ver =~ s/\.lsb\d\d?//; # remove .lsb(2|20|3)
                debug_inform("$package_name: $package_ver");
                if ( $package_name eq "lsb-tet3-lite" ) { $package_name = "lsb-tet3"; }
                if ( $package_name eq "qm" ) { $package_name = "lsb-qm"; }
                if ( $package_name eq "qmtest-libstdcpp" ) { $package_name = "lsb-test-libstdcpp"; }
                if ( $package_name eq "lsb-runtime-test" ) { $package_name = "lsb-test-core"; }
                $result{$package_name} = $package_ver;
            }
        }
    }
    else {
        # Unknown package manager
        fail( "packages_info: Unknown package manager."
             ." ($host_package_manager FOR $host_os)"
            ); 
    }
    
    return \%result;
}
#------------------------------------------------------------------------
sub package_file_location {
    my $pkg_name = shift;
    my $filename = shift;
    
    my $path = "";
    my $true_name = "";
    unless (package_info ($pkg_name, \$true_name)) { return ""; }
    if ( $host_package_manager eq "RPM" ) {
        $path = capture ("rpm -ql $true_name | grep '/$filename\$'");
    }
    elsif ( $host_package_manager eq "dpkg" ) {
        $path = capture("dpkg -L $true_name | grep '/$filename\$'");
    }
    $path =~ s/\/[^\/]*$//;
    return $path;
}
#------------------------------------------------------------------------
    
# Installs the specified package file.
sub install_package
{
    my $package_file = shift;
    
    # Check that package_file exists
    unless(-f $package_file) {
        inform ("File doesn't exist: $package_file");
        return 1;
    }
    
    # Install it.
    inform ("Installing: $package_file");
    
    if( $host_package_manager eq "RPM" )
    {
        # Use RPM
        cmd ("rpm -i $package_file"); 
    }
    elsif( $host_package_manager eq "dpkg" )
    {
        # Use 'alien' and 'dpkg'.
        my $convert_dir = '/tmp/lsb-dtk-convert';
        if ( -d $convert_dir ) {  cmd ("rm -rf $convert_dir");  }
        mkdir ($convert_dir);
        my $resp = capture ("cd '$convert_dir' "
            ." && alien -d -c -k $package_file 2>/dev/null"); 
        debug_inform("alien: $resp");
        $resp =~ s/ .*//;
        chomp $resp;
        if ($resp) {
            cmd("dpkg -i $convert_dir/$resp");
        }
    }
    else {
        fail( "install_package: Unknown package manager."
             ." ($host_package_manager, $host_os)"
        ); 
    }
    
    return 0;
}
#----------------------------------------------------------------------------

# Uninstalls the specified package if it is installed.
sub remove_package
{
    my $package_name = shift;

    if ($package_name eq "lsb-tet3") {
        remove_package("lsb-tet3-lite");
    }
    if ($package_name eq "lsb-qm") {
        remove_package("qm");
    }
    elsif ($package_name eq "lsb-test-libstdcpp") {
        remove_package("qmtest_libstdcpp");
    }
    elsif ($package_name eq "lsb-test-core") {
        remove_package("lsb-runtime-test");
    }
    
    # Check whether the package is installed.
    if (!package_info($package_name)) {
        # The package seems to be not installed.
        return 0;
    }
    
    # The package seems to be installed, uninstall it.

    # Failure to uninstall package is (mostly) not fatal;
    # either the original version will be used, or a cleanup will not
    # be complete. Try to continue.
    if( $host_package_manager eq "RPM" )
    {
        # Use RPM
        cmd ("rpm -e --allmatches --nodeps $package_name"); 
    }
    elsif( $host_package_manager eq "dpkg" ) {
        # Use 'alien' and 'dpkg'.
        my $response = capture("dpkg -P $package_name 2>&1"); 
        
        if( $? != 0 || $response =~ /dpkg:/ ) {
            # Package in qmtest_libstdcpp-*.rpm can be named 
            # like qmtest-libstdcpp.
            $package_name =~ s/_/-/g; # qmtest_libstdcpp case
            cmd ("dpkg -P $package_name");
        }
    }
    else {
        fail( "remove_package: Unknown package manager."
             ." ($host_package_manager, $host_os)"
        ); 
    }
    
    return 0;
}
#------------------------------------------------------------------------

sub login_ftp {
    # Check FTP perl module
    if( cmd ("perl -MNet::FTP -e 1 >/dev/null 2>&1") != 0) {
            fail( "Perl module Net::FTP is required." );
    }
    
    # Open ftp
    require Net::FTP;
    
    debug_inform ("FTP init... $FTP_srv");
    $ftp = Net::FTP->new($FTP_srv, Debug => 0, Passive => 1);
    unless ( $ftp ) { fail("FTP init failed. Message: $@"); }
    
    # Log in
    debug_inform ("FTP log in... $FTP_srv");
    unless ( $ftp->login() ) { 
        fail("FTP log in (anonymous) failed. Message: ".$ftp->message); 
    };
}
#------------------------------------------------------------------------

# Reads FTP listing from a file.
sub read_listing_from_file {
    my $filename = shift;
    my @listing = (); #result; 
    
    open FILE, $filename or die "Can't open $filename";
    
    while (<FILE>) {
        chomp;
        if ( /^\s*#/ ) {
            if ( /^#TIMESTAMP:(\d{2})-(\d{2})-(\d{4})\s+(\d{2}):(\d{2})$/ ) {
                # Timestamp helps to determine when the cache expires.
                my $timestamp = mktime(0, $5, $4, $1, $2-1, $3-1900);
                if ($timestamp < $oldest_timestamp) { $oldest_timestamp = $timestamp; }
            }
            elsif ( /^#PWD:(.*)$/ ) {
                my $path = $1;
                # One could check the path here
            }
            else {
                debug_inform("Wrong syntax in file $filename");
            }
        }
        else {
            # It is a regular DIR line
            push (@listing, $_);
        }
    }
    close FILE;
    return \@listing;
}
#------------------------------------------------------------------------

# Returns file listing of a given dir on the ftp server.
sub list_ftp
{
    my $pref = shift; # Prefix ("pub/lsb" actually)
    my $src = shift; # path to the dir, which listing should be taken
    
    my $source = $pref.$src;
    
    my @ftp_files = (); # results: @ftp_files[] -> @($source, $filename, $filesize)
    
    my $listing = ();
    
    # Prepare corresponding local directory
    unless (-d "$package_dir/$src" ) {
        cmd ("mkdir -p \"$package_dir/$src\"");
    }
    
    debug_inform ("> SRC $src");
    
    unless( $list_refresh || $force_dl ) {
        # Read cached listing
        if ( -f $package_dir.$src."/ftp_list" ) {
            $listing = read_listing_from_file ($package_dir.$src."/ftp_list");
        }
    } 
    if(!$listing && $download_packages ) { 
        # Refresh the listing.
        if ("-f \"".$package_dir.$src."/ftp_list\"") {
            cmd("rm -f \"".$package_dir.$src."/ftp_list\"");
        }
        unless ($ftp) { login_ftp(); }
        $listing = $ftp->dir($source);
    }
    
    unless($listing) { 
        debug_inform ("WRONG DIR");
        $oldest_timestamp = 0;
        # Return empty list
        return @ftp_files;
    }
    
    # Save listing
    unless ( -f $package_dir.$src."/ftp_list" ) {
        my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ();
        write_string_as_file ($package_dir.$src."/ftp_list", 
            "#TIMESTAMP:"
            .sprintf ("%02d-%02d-%04d %02d:%02d\n", $mday, ($mon + 1), 
            ($year + 1900), $hour, $min )
           ."#PWD:".$source."\n"
           .(join "\n", @$listing)
           );
    }
    
    unless(@$listing) { 
            debug_inform ("< Empty");
            # Return empty list
            return @ftp_files;
    }

    # Parse ftp files listing
    foreach my $line (@$listing) {
        # -rwxr-xr-x 1 user group size date time filename.ext
        if($line =~ m/^([^ ]+)\s+[^ ]+\s+[^ ]+\s+[^ ]+\s+([^ ]+)\s+[^ ]+\s+[^ ]+\s+[^ ]+\s+([^ ]+)/) {
            my $attrs = $1;
            my $filesize = $2;
            my $filename = $3;
            
            # If it is an ordinary file or a symlink
            if($attrs =~ /^[-l]/ ) {
                my @line = ($source, $filename, $filesize);
                push @ftp_files, \@line;
            }
        }
        else { 
            debug_inform("Wrong syntax in ftp listing: $line");
        }
    }
    debug_inform ("< Ok");

    return @ftp_files;
}
#----------------------------------------------------------------------------

# Adds file to list of all available files
sub add_file_to_list {
    my ($filename, $package_name, $pkg_ver, $mode, $src, $size, $lsb_ver, $arch) = @_;
    
    #debug_inform("+ $mode $lsb_ver $arch $package_name $filename");
    
    # Extract package_name, version and architecture from the filename
    if ( $filename =~ m/^(([^\d][\w_]*-)*)(\d.*)\.(\w+)\.rpm$/ ) {
        my $new_package_name = $1;
        my $new_package_ver = $3;
        my $new_arch = $4;
        if ( !$package_name ) { 
            $package_name = $new_package_name; 
            $package_name =~ s/-$//g;
            
            # tricky case with the lsb-tet3-lite beta for LSB 3.0
            if ( $package_name eq "lsb-tet3-lite" ) { $package_name = "lsb-tet3"; }
            if ( $package_name eq "qm" ) { $package_name = "lsb-qm"; }
            if ( $package_name eq "qmtest_libstdcpp" ) { $package_name = "lsb-test-libstdcpp"; }
            if ( $package_name eq "lsb-runtime-test" ) { $package_name = "lsb-test-core"; }
        }
        if ( !$pkg_ver ) { $pkg_ver = $new_package_ver; }
        if ( !$arch ) { 
            if ( $new_arch eq 'noarch' ) {
                $arch = 'all';
            } else {
                $arch = guess_platform($new_arch); 
            }
        }    
    } else {
        if( $filename =~ m/\.rpm$/ ) {
            debug_inform ( "Bad file name: $filename." );
        }
        if ( !$package_name ) { $package_name = $filename; }
    }
    
    $pkg_ver =~ s/\.lsb\d\d?$//; # remove .lsb(2|20|3)
    
    foreach my $lsb_ver_i ( @all_lsb_versions ) {
        if ( $lsb_ver && ($lsb_ver ne "all") && ($lsb_ver ne $lsb_ver_i) ) {next;}
        foreach my $arch_i ( @all_platforms ) {
            if ( $arch && ($arch ne "all") && ($arch ne $arch_i) ) {next;}

            my $file  = { FILENAME => $filename, 
                          PACKAGE  => $package_name,
                          PKG_VER  => $pkg_ver,
                          MODE     => $mode,
                          SRC      => $src,
                          SIZE     => $size,
                          LSB_VER  => $lsb_ver_i,
                          ARCH     => $arch_i,
                        };
            if ( (!defined $s_files->{$mode}->{$lsb_ver_i}->{$arch_i}->{$package_name})
              || !($s_files->{$mode}->{$lsb_ver_i}->{$arch_i}->{$package_name}->{PKG_VER} gt $file->{PKG_VER})
            ){
                $s_files->{$mode}->{$lsb_ver_i}->{$arch_i}->{$package_name} = $file;
                #debug_inform("+ $mode $lsb_ver_i $arch_i $package_name $filename");
            } 
            
            if ( !exists($files{"$package_name.$pkg_ver.$arch_i"}) ) {
                $files{"$package_name.$pkg_ver.$arch_i"} = $file;
            }
        }
    }
}
#----------------------------------------------------------------------------

# Gather a set of sources from all tests.
sub gather_sources
{
    my $sources = {}; # Resulting set
    
    my %src_vals = ();
    foreach my $test_obj (values %tests) {
        # release sources
        foreach my $src ( split " ", $test_obj->sources) {
            next unless $src;
            $src_vals{$src} = 1;
        }
    }
    foreach my $lsb_ver ( @all_lsb_versions ) {
        next if $lsb_ver eq 'snapshot';
        
        foreach my $src ( keys %src_vals ) {
            if ($src eq "runtime") {
                my $runtime_url = "/test_suites/released-"
                                  .$lsb_ver."/binary/runtime";
                $sources->{"r"}->{$lsb_ver}->{"all"}->{$runtime_url} = 1;
            }
            elsif ($src eq "appbat") {
                foreach my $arch ( @all_platforms ) {
                    next unless is_backward_compatible_platform($arch);
                    my $appbat_url = "/app-battery/released-".$lsb_ver."/".$arch;
                    $sources->{"r"}->{$lsb_ver}->{$arch}->{$appbat_url} = 1;
                }
            }
            elsif ($src eq "lsbdev") {
                foreach my $arch ( @all_platforms ) {
                    next unless is_backward_compatible_platform($arch);
                    my $lsbdev_url = "/lsbdev/released-".$lsb_ver."/binary/".$arch;
                    $sources->{"r"}->{$lsb_ver}->{$arch}->{$lsbdev_url} = 1;
                }
            }
            elsif ($src eq "test_data") {
                my $data_url = "/app-battery/tests";
                $sources->{"r"}->{$lsb_ver}->{"all"}->{$data_url} = 1;
            }
            else {
                $sources->{"r"}->{$lsb_ver}->{"all"}->{$src} = 1;
            }
        } # end of foreach source names
    } # end of foreach lsb_ver
    
    my %beta_src_vals = ();
    foreach my $test_obj (values %tests) {
        # release sources
        foreach my $src ( split " ", $test_obj->beta_sources) {
            next unless $src;
            $beta_src_vals{$src} = 1;
        }
    }
    foreach my $lsb_ver ( @all_lsb_versions ) {
        next if $lsb_ver eq 'snapshot';
        
        foreach my $src ( keys %beta_src_vals ) {
            if ($src eq "runtime") {
                my $beta_runtime_url = "/test_suites/beta/binary/runtime";
                $sources->{"b"}->{$lsb_ver}->{"all"}->{$beta_runtime_url} = 1;
            }
            elsif ($src eq "appbat") {
                foreach my $arch ( @all_platforms ) {
                    next unless is_backward_compatible_platform($arch);
                    my $beta_appbat_url = "/app-battery/beta/".$arch;
                    $sources->{"b"}->{$lsb_ver}->{$arch}->{$beta_appbat_url} = 1;
                }
            }
            elsif ($src eq "test_data") {
                my $data_url = "/app-battery/tests";
                $sources->{"r"}->{$lsb_ver}->{"all"}->{$data_url} = 1;
            }
            else {
                $sources->{"b"}->{$lsb_ver}->{"all"}->{$src} = 1;
            }
        } # end of foreach beta source names
    } # end of foreach lsb_ver
    
    my %snapshot_src_vals = ();
    foreach my $test_obj (values %tests) {
        # release sources
        foreach my $src ( split " ", $test_obj->snapshot_sources) {
            next unless $src;
            $snapshot_src_vals{$src} = 1;
        }
    }
    foreach my $src ( keys %snapshot_src_vals ) {
        if ($src eq "appbat") {
            my $snapshot_appbat_url = "/snapshots/appbat";
            $sources->{"s"}->{"snapshot"}->{"ia32"}->{$snapshot_appbat_url} = 1;
        }
        elsif ($src eq "test_data") {
            my $data_url = "/snapshots/appbat/tests";
            $sources->{"s"}->{"snapshot"}->{"ia32"}->{$data_url} = 1;
            $data_url = "/app-battery/tests";
            $sources->{"s"}->{"snapshot"}->{"ia32"}->{$data_url} = 1;
        }
        else {
            $sources->{"s"}->{"snapshot"}->{"ia32"}->{$src} = 1;
        }
    } # end of foreach snapshot source names
    return $sources;
}
#----------------------------------------------------------------------------

# Generates a list of all available files for each certain platform and mode
sub list_files {
    my $max_mode = shift;
    debug_inform ("Max mode: $max_mode");
    
    unless ( $force_reinstall ) {
        ## Register installed packages
        # Collect all required packages
        my %tmp_hash = ();
        foreach my $test_obj ( values %tests ) {
            foreach my $package_name (split " ", $test_obj->{PACKAGES}) {
                $tmp_hash{$package_name} = 1;
            }
        }
        
        my $result = packages_info (keys %tmp_hash);
        
        foreach my $package_name (keys %$result) {
            my $info = $result->{$package_name};
            if ( $info ) {
                # add installed packages to list
                add_file_to_list("$package_name-$info.$host_platform.rpm",
                    $package_name, $info, "p", "", 0, "all", $host_platform );
            }
        }
    }
    #----------
    
    ## Register packages in package_dir
    
    my $sources = gather_sources();
    
    foreach my $rb ("r","b","s") {
        my $mode = $rb;
        if ($rb eq "r") { $mode = "p"; }
        if (  ( ($max_mode eq "s") || ($rb eq "s") ) && ($max_mode ne $rb)  ) { next; }
        
        foreach my $lsb_ver ( keys %{$sources->{$rb}} ) {
            foreach my $arch ( keys %{$sources->{$rb}->{$lsb_ver}} ) {
                foreach my $src ( keys %{$sources->{$rb}->{$lsb_ver}->{$arch}} ) {
                    my $dir = $package_dir.$src;
                    unless (opendir DIR, $dir) { next; }
                    
                    # Read directory for files and dirs
                    my @dir_list = readdir(DIR);
                    
                    foreach my $filename(@dir_list) {
                        # skip hidden files, . and ..
                        next if $filename =~ /^\./;
                        
                        # '-f' checks it is a file
                        if (-f "$dir/$filename") {
                            add_file_to_list($filename, "","",$mode, $src, 0, $lsb_ver, "");
                        }
                    }
                }
            }
        }
    }
    #----------
    
    ## Look for packages on the FTP
    if ( $max_mode =~ /[abs]/ ) 
    {
        # Request FTP files list
        
        # Count up sources
        my $number = 0;
        foreach my $rb (keys %$sources) {
            if ($rb eq "b") { next if ($max_mode ne "b"); }
            if (  ( ($max_mode eq "s") || ($rb eq "s") ) && ($max_mode ne $rb)  ) { next; }
            
            foreach my $lsb_ver ( keys %{$sources->{$rb}} ) {
                foreach my $arch ( keys %{$sources->{$rb}->{$lsb_ver}} ) {
                    $number += keys %{$sources->{$rb}->{$lsb_ver}->{$arch}};
                }
            }
        }
        
        $oldest_timestamp = time();
        
        # List dirs
        my $count = 0;
        foreach my $rb (keys %$sources) {
            my $mode = $rb;
            if ($mode eq "r") { $mode = "a"; }
            if ($rb eq "b") { next if ($max_mode ne "b"); }
            if (  ( ($max_mode eq "s") || ($rb eq "s") ) && ($max_mode ne $rb)  ) { next; }
            
            foreach my $lsb_ver ( keys %{$sources->{$rb}} ) {
                foreach my $arch ( keys %{$sources->{$rb}->{$lsb_ver}} ) {
                    foreach my $src ( keys %{$sources->{$rb}->{$lsb_ver}->{$arch}} ) {
                        
                        my @list = list_ftp("pub/lsb", $src);
                        $count++;
                        if ( $list_refresh || $force_dl || ($verbose_lvl >= 2) ) {
                            print "#PROGRESS: $count/$number\n";
                            print "#\x08 ", "\033[2A","\n";
                        }
                        
                        foreach my $line ( @list ) {
                            my ($source, $filename, $filesize) = @$line;
                            #debug_inform("  $lsb_ver ($mode) $filename");
                            add_file_to_list( $filename,"","",$mode, $src,
                                              $filesize, $lsb_ver, "" );
                        }
                    }
                }
            }
        }
        
        debug_inform ( 'Done' . (' ' x 10) );
        
        # Close FTP connction
        if ( $ftp ) {
            $ftp->quit;
            $ftp = undef;
        }
    }
    
    #----------
    
    # Choose right versions of required files for each test
    foreach my $lsb_ver ( @all_lsb_versions ) {
        foreach my $arch ( @all_platforms ) {
            next unless is_backward_compatible_platform($arch);
            foreach my $test_obj (values %tests) {
                my $test_name = $test_obj->name;
                foreach my $mode ("p","a","b","s") {
                    my @files_to_search = split " ", $test_obj->files($lsb_ver);
                    my @packages_to_search = split " ", $test_obj->packages($lsb_ver);
                    foreach my $filename_search (@files_to_search, @packages_to_search) {
                        # max_mode is the mode of the last (main) package
                        $max_mode = $mode;
                        my $file = $s_files->{$mode}->{$lsb_ver}->{$arch}->{$filename_search};
                        if ( ($mode eq "b") && !$file) {
                            $file = $s_files->{"a"}->{$lsb_ver}->{$arch}->{$filename_search};
                            $max_mode = "a";
                        }
                        if ( !$file) {
                            $file = $s_files->{"p"}->{$lsb_ver}->{$arch}->{$filename_search};
                            $max_mode = "p";
                        }
                        unless ($file) {
                            # Can't find any suitable file, so this test will be unavailable
                            delete $test_info->{$lsb_ver}->{$arch}->{$test_name}->{$mode};
							#debug_inform("- $mode $lsb_ver $arch $test_name : $filename_search");
                            last;
                        }                        

                        # select the best file
                        my $key = $file->{PACKAGE}.".".$file->{PKG_VER}.".".$file->{ARCH};
                        $file = $files{$key};
                        
                        $test_info->{$lsb_ver}->{$arch}->{$test_name}->{$mode}->{FILES}->{$filename_search}=$file;
                        $test_info->{$lsb_ver}->{$arch}->{$test_name}->{$mode}->{VERSION}=$file->{PKG_VER};
                    }
                    if ( defined $test_info->{$lsb_ver}->{$arch}->{$test_name}->{$mode} ) {
                        unless ($max_mode eq $mode) {
                            # This mode isn't present.
                            delete $test_info->{$lsb_ver}->{$arch}->{$test_name}->{$mode};
                            #debug_inform("- $mode $lsb_ver $arch $test_name: (only '$max_mode')");
                        }
                    }
                }
            }
        }
    }
}
#----------------------------------------------------------------------------

# Produce the text listing of required files
sub print_files_list {
    
    print <<TEXT;
# This is a file with the list of all available test modules.
# Syntax:
# ARCHITECTURE: <arch>
#     Current machine architecture
# FILE_SIZE: <file> <size>
#     Size in bytes needed to be downloaded for the file specified
# MODULE_TIME: <module> <time>
#     Time in special units (seconds for a reference machine) needed for the
#     specified test suite to run. These units are then translated into seconds.
# <lsb-version> <arch> <module> <location> <version> <files>
#     <lsb-version> : LSB version for which the information is given
#     <arch>        : machine architecture that the module is for 
#                     (i386, ia64 etc.)
#     <module>      : module name
#     <location>    : location of the test suites ('p' (present) stands for 
#                     localhost, 'a' (available) for FTP server, 'b' (beta) 
#                     for 'beta' subdirectory on the FTP)
#     <version>     : version of the test suite
#     <files>       : comma-separated list of files needed for the test suite
TEXT

    print "ARCHITECTURE: $host_platform\n";
    print "LSB_VERSION: $lsb_version\n";
    
    if (  (defined $oldest_timestamp) && ( $oldest_timestamp < time() )
        && ( abs($oldest_timestamp - time()) > $expire_period )  
        ) {
            if ( $oldest_timestamp > 0 ) {
                my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) = localtime ($oldest_timestamp + $expire_period);
                print "EXPIRED: ".sprintf ("%02d-%02d-%04d %02d:%02d\n", $mday, ($mon + 1), 
                    ($year + 1900), $hour, $min )."\n";
            } else {
                print "EXPIRED: undef\n";
            }
    }
    
    my %tmp_hash = ();
    foreach my $lsb_ver ( keys %$test_info ) { 
        foreach my $platform ( keys %{$test_info->{$lsb_ver}} ) { 
            foreach my $test_name ( keys %{$test_info->{$lsb_ver}->{$platform}} ) {
                my $tmp = $test_info->{$lsb_ver}->{$platform}->{$test_name};
                foreach my $mode (keys %$tmp)
                {
                    next unless ($tmp->{$mode});
                    
                    foreach my $file (values %{$tmp->{$mode}->{FILES}}) {
                        my $filename = $file->{FILENAME};
                        $tmp_hash{$filename} = $file;
                    }
                }
            }
        }
    }
    foreach my $filename (keys %tmp_hash) {
        my $file = $tmp_hash{$filename};
        print "FILE_SIZE: ".$file->{FILENAME}."\t".$file->{SIZE}.($verbose_lvl>1?"\t".$file->{SRC}:"")."\n";
    }
    
    # Print required files for each test, platform and mode
    foreach my $lsb_ver ( keys %$test_info ) { 
        foreach my $platform (keys %{$test_info->{$lsb_ver}}) {
            foreach my $test_name ( keys %{$test_info->{$lsb_ver}->{$platform}} )
            {
                my $tmp = $test_info->{$lsb_ver}->{$platform}->{$test_name};
                my $display_name = $tests{$test_name}->{DISPLAYNAME};
                foreach my $mode (keys %$tmp)
                {
                    next unless ($tmp->{$mode});
                    print "$lsb_ver\t$platform\t$test_name\t$display_name";
                    my $version = $tmp->{$mode}->{VERSION};
                    print "\t$mode\t$version";
                    my (@tmp1, @tmp2, @tmp3);
                    foreach my $package_name ( keys %{$tmp->{$mode}->{FILES}} ) {
                        push @tmp1, $tmp->{$mode}->{FILES}->{$package_name}->{PKG_VER};
                        if ( $tmp->{"p"} && $tmp->{"p"}->{VERSION} ) {
                            push @tmp2, $tmp->{"p"}->{FILES}->{$package_name}->{PKG_VER};
                        }
                        if ( $tmp->{"a"} && $tmp->{"a"}->{VERSION} ) {
                            push @tmp3, $tmp->{"a"}->{FILES}->{$package_name}->{PKG_VER};
                        }
                    }
                    if( $mode ne "p" && 
                        (
                            $tmp->{"p"} && @tmp2
                            && ( $tmp->{$mode}->{VERSION} eq $tmp->{"p"}->{VERSION} )
                            && ( "@tmp1" ne "@tmp2" )
                                ||
                            $tmp->{"a"} && @tmp3
                            && ( $tmp->{$mode}->{VERSION} eq $tmp->{"a"}->{VERSION} )
                            && ( "@tmp1" ne "@tmp3" )
                        )
                       )
                       { print "+"; }
                    @tmp1 = ();
                    foreach my $package_name ( keys %{$tmp->{$mode}->{FILES}} ) {
                        #if ( $tmp->{$mode}->{FILES}->{$package_name}->{SRC} ) {
                        {
                            push @tmp1, $tmp->{$mode}->{FILES}->{$package_name}->{FILENAME};
                        }
                    }
                    print "\t", (join ", ", @tmp1);
                    print "\n";
                }
            }
        }
    }
}
#------------------------------------------------------------------------

sub inquire_for_files_list {
    
    my %to_ajust_mode = ();
    
    # Determine max_mode
    my $max_mode = "p"; 
    foreach my $test_name (@tests_to_run) {
        my $mode = $test_mode{$test_name};
        
        unless ($mode) {
            $mode = "p";
            
            if ( $download_packages ) { 
                if ( $try_beta_packages ) {
                    $mode = "b";
                } else {
                    $mode = "a";
                }
            }
            if ( $lsb_version eq "snapshot" ) { $mode = "s"; }
            
            $test_mode{$test_name} = $mode;
            $to_ajust_mode{$test_name} = 1;
        }

        if ( ($mode eq "a") && ($max_mode eq "p") ) {
            $max_mode = "a"
        } elsif ( ($mode eq "b") && ($max_mode ne "b") ) {
            $max_mode = "b";
        } elsif ( ($mode eq "s") && ($max_mode ne "s") ) {
            $max_mode = "s";
        }
    }

    debug_inform ("Max mode: $max_mode");
    list_files ($max_mode);
    if ( $verbose_lvl >= 2 ) {
        print_files_list ();
    }
    
    unless ( $certification_mode ) {
        # Adjust modes given automaticaly
        foreach my $test_name ( keys %to_ajust_mode ) {
            if( ($test_mode{$test_name} eq "b") 
                    && !$test_info->{$lsb_version}->{$host_platform}->{$test_name}->{"b"}  ) {
                $test_mode{$test_name} = "a";
                debug_inform("$test_name mode set to 'a'");
            }
            if( ($test_mode{$test_name} eq "a") 
                    && !$test_info->{$lsb_version}->{$host_platform}->{$test_name}->{"a"}  ) {
                $test_mode{$test_name} = "p";
                debug_inform("$test_name mode set to 'p'");
            }
            if( ($test_mode{$test_name} eq "s") 
                    && !$test_info->{$lsb_version}->{$host_platform}->{$test_name}->{"s"}  ) {
                $test_mode{$test_name} = "p";
                debug_inform("$test_name mode set to 'p'");
            }
        }
    }
  
    # Check tests availability
    my $i = 0;
    my @tests_to_run_copy = @tests_to_run;
    foreach my $test_name (@tests_to_run_copy) {
        if( !defined($test_info->{$lsb_version}->{$host_platform}->{$test_name}->{$test_mode{$test_name}} ) ) {
            unless($certification_mode) {
                inform ("Test $test_name is not available with this settings."
                .( ($download_packages && $try_beta_packages)
                        ?"":" You could try to run tests with -D and --beta parameters.")  );
            }
            # Remove it from @tests_to_run
            splice (@tests_to_run, $i, 1);
        } else {
            $i++;
        }
    }
}
#------------------------------------------------------------------------

# Download required packages.
sub prepare_test {
    
    my $test_obj = shift;
    my $test_name = $test_obj->name;
    
    unless ( chdir("$package_dir") ) {
        inform ("Can't access $package_dir");
        return 1;
    }
    
    my $mode = $test_mode{$test_name};
    
    my $tmp = $test_info->{$lsb_version}->{$host_platform}->{$test_name}->{$mode};
    
    unless ($tmp) { return 1 };
    
    # Download and install required packages
    foreach my $package_name ( split " ", $test_obj->packages($lsb_version) )
    {
        my $file = $tmp->{FILES}->{$package_name};
        my $info = "";
        if ( ! $force_reinstall ) {
            my $true_package_name = "";
            $info = package_info($package_name, \$true_package_name);
            
            debug_inform("$true_package_name: $info");
        }
        
        # If installed package version doesn't equal to required version
        # or if package not installed ( $info = "" )
        if( ($info ne $file->{PKG_VER}) || $force_reinstall ) {
            # Download required files if -D option specified 
            # and file doesn't actually exists in package_dir.
            my $filedest = $file->{SRC}."/".$file->{FILENAME};
            my $filesrc = $file->{SRC};
            my $filename = $file->{FILENAME};
            if ( $download_packages && 
                    (!(-f $package_dir.$filedest) || $force_dl) 
                )
            {
                if ( !$filesrc ) {
                    inform ("Can't find package that is considered to be installed: $filedest");
                    return 1;
                }
                # To avoid partially downloaded (broken) files
                # download the file into /tmp directory, then move it
                # to $package_dir.
                                
                if ( (-f $package_dir.$filedest) && $force_dl ) {
                    # Copy existing file to /tmp to let wget make
                    # a decision whether to download a newer copy of file
                    # or not.
                    cmd("cp -f -p '".$package_dir.$filedest."' /tmp");
                }
                
                # Use wget utility to download files
                my $result = cmd( "cd /tmp && $wget '"
                                  .$FTP_lsb.$filedest
                                  ."' && mv -f '$filename' '"
                                  .$package_dir.$filedest."'" );
                if ( $result != 0 ) {
                    inform ("$test_name: Sorry, wget failed to download $filedest.");
                    return 1;
                }
            }
        }
    }
    
    foreach my $file_name ( split " ", $test_obj->files($lsb_version) )
    {
        my $file = $tmp->{FILES}->{$file_name};
        my $filedest = $file->{SRC}."/".$file->{FILENAME};
        my $filesrc = $file->{SRC};
        my $filename = $file->{FILENAME};
        if ( $download_packages && 
                ( !(-f $package_dir.$filedest) || $force_dl || $force_reinstall)  
            )
        {
            if ( !$filesrc ) {
                inform ("Wrong source defined for file: $filedest");
                return 1;
            }
            
            if ( (-f $package_dir.$filedest) && $force_dl ) {
                cmd("cp -f -p '".$package_dir.$filedest."' /tmp");
            }
            
            # Use wget utility to download files
            my $result = cmd( "cd /tmp && $wget '"
                              .$FTP_lsb.$filedest
                              ."' && mv -f '$filename' '"
                              .$package_dir.$filedest."'" );
            if ( $result != 0 ) {
                inform ("$test_name: wget failed to download $filedest.");
                return 1;
            }
        }
    }

    return 0;
}

# Install required packages
sub install_test {
    
    my $test_obj = shift;
    my $test_name = $test_obj->name;
    
    unless(-d "$results_dir/host_info") {
        cmd("mkdir -p $results_dir/host_info");
    }
    append_string_to_file("$results_dir/host_info/package_versions"
                          , "TEST: $test_name\n");
    
    unless ( chdir("$package_dir") ) {
        inform ("Can't access $package_dir");
        return 1;
    }
    
    my $mode = $test_mode{$test_name};
    
    my $tmp = $test_info->{$lsb_version}->{$host_platform}->{$test_name}->{$mode};
    
    unless ($tmp) {
        debug_inform("No information about test packages.");
        return 1; 
    }
    
    $test_obj->{RUN_VERSION} = $tmp->{VERSION};
    
    # Download and install required packages
    foreach my $package_name ( split " ", $test_obj->packages($lsb_version) )
    {
        my $file = $tmp->{FILES}->{$package_name};
        my $true_package_name = "";
        my $info = package_info($package_name, \$true_package_name);
        
        debug_inform("$true_package_name: $info");
        
        # If installed package version doesn't equal to required version
        # or if package not installed ( $info = "" )
        if( ($info ne $file->{PKG_VER}) || $force_reinstall ) {
	        if ( $info ne $file->{PKG_VER} ) {
	            debug_inform("Installed version '$info' doesn't coincide with selected '$file->{PKG_VER}'.");
            }
            # Download required files if -D option specified 
            # and file doesn't actually exists in package_dir.
            my $filedest = $file->{SRC}."/".$file->{FILENAME};
            
            if ( $info ) {
                remove_package($package_name);
            }
            
            install_package($package_dir.$filedest);
            
            $info = package_info($package_name, \$true_package_name);
        }
        
        # Check whether the package was installed successfully.
        if ( $info ne $file->{PKG_VER} )
        {
            inform ("Package ".$file->{SRC}."/".$file->{FILENAME}." ($true_package_name ".$file->{PKG_VER}.") was not correctly installed.");
            return 1;
        }
        
        append_string_to_file("$results_dir/host_info/package_versions"
                              , "  ".$package_name."\t".$info."\n");
    }
}

1; # return value

# Closes the ftp connection if the execution was interrupted
sub END {
    if ( $ftp ) {
        $ftp->quit;
    }
}
